from cStringIO import StringIO
from functools import wraps
import hashlib
import json
import re
import logging
from itertools import islice, chain

from requests import Session as RequestSession, Response as RequestResponse
from requests.structures import CaseInsensitiveDict

from django.conf import settings
from django.core.cache import cache
from django.core.exceptions import PermissionDenied, ObjectDoesNotExist
from django.core.serializers.json import DjangoJSONEncoder
from django.http import HttpRequest, HttpResponse, HttpResponseBadRequest
from django.template import Template, Context
from django.utils.decorators import available_attrs
from django.views.generic.detail import SingleObjectMixin
from django.contrib.contenttypes.models import ContentType
from django.db import transaction
from zeep import Client
from zeep.cache import InMemoryCache
from zeep.transports import Transport

from passerelle.base.context_processors import template_vars
from passerelle.base.models import ApiUser, AccessRight, BaseResource
from passerelle.base.signature import check_query, check_url

from .jsonresponse import to_json

def get_template_vars():
    """
    returns the template vars as dict, to be used in apps code
    """
    from django.http import HttpRequest
    return template_vars(HttpRequest())

def render_template_vars(value):
    """
    renders the template vars in a string
    """
    template = Template(value)
    return template.render(Context(get_template_vars()))


def response_for_json(request, data):
    response = HttpResponse(content_type='application/json')
    json_str = json.dumps(data)
    for variable in ('jsonpCallback', 'callback'):
        if variable in request.GET:
            identifier = request.GET[variable]
            if not re.match(r'^[$A-Za-z_][0-9A-Za-z_$]*$', identifier):
                return HttpResponseBadRequest('invalid JSONP callback name')
            json_str = '%s(%s);' % (identifier, json_str)
            response['Content-Type'] = 'application/javascript'
            break
    response.write(json_str)
    return response


def get_request_users(request):
    users = []

    users.extend(ApiUser.objects.filter(keytype=''))

    if 'orig' in request.GET and 'signature' in request.GET:
        orig = request.GET['orig']
        query = request.META['QUERY_STRING']
        signature_users = ApiUser.objects.filter(keytype='SIGN', username=orig)
        for signature_user in signature_users:
            if check_query(query, signature_user.key):
                users.append(signature_user)

    elif 'apikey' in request.GET:
        users.extend(ApiUser.objects.filter(keytype='API',
            key=request.GET['apikey']))

    elif request.META.has_key('HTTP_AUTHORIZATION'):
        (scheme, param) = request.META['HTTP_AUTHORIZATION'].split(' ',1)
        if scheme.lower() == 'basic':
            username, password = param.strip().decode('base64').split(':',1)
            users.extend(ApiUser.objects.filter(keytype='SIGN',
                    username=username, key=password))

    def ip_match(ip, match):
        if not ip:
            return True
        if ip == match:
            return True
        return False

    users = [x for x in users if ip_match(x.ipsource, request.META.get('REMOTE_ADDR'))]
    return users

def get_trusted_services():
    '''
    All services in settings.KNOWN_SERVICES are "trusted"
    '''
    trusted_services = []
    for service_type in getattr(settings, 'KNOWN_SERVICES', {}):
        for slug, service in settings.KNOWN_SERVICES[service_type].iteritems():
            if service.get('secret') and service.get('verif_orig'):
                trusted_service = service.copy()
                trusted_service['service_type'] = service_type
                trusted_service['slug'] = slug
                trusted_services.append(trusted_service)
    return trusted_services


def is_trusted(request):
    '''
    True if query-string is signed by a trusted service (see get_trusted_services() above)
    '''
    if not request.GET.get('orig') or not request.GET.get('signature'):
        return False
    full_path = request.get_full_path()
    for service in get_trusted_services():
        if (service.get('verif_orig') == request.GET['orig']
                and service.get('secret')
                and check_url(full_path, service['secret'])):
            return True
    return False


def is_authorized(request, obj, perm):
    if is_trusted(request):
        return True
    resource_type = ContentType.objects.get_for_model(obj)
    rights = AccessRight.objects.filter(resource_type=resource_type,
            resource_pk=obj.id, codename=perm)
    users = [x.apiuser for x in rights]
    return set(users).intersection(get_request_users(request))


def protected_api(perm):
    def decorator(view_func):
        @wraps(view_func, assigned=available_attrs(view_func))
        def _wrapped_view(instance, request, *args, **kwargs):
            if not isinstance(instance, SingleObjectMixin):
                raise Exception("protected_api must be applied on a method of a class based view")
            obj = instance.get_object()
            if not is_authorized(request, obj, perm):
                raise PermissionDenied()
            return view_func(instance, request, *args, **kwargs)
        return _wrapped_view
    return decorator


def content_type_match(ctype):
    content_types = settings.LOGGED_CONTENT_TYPES_MESSAGES
    if not ctype:
        return False
    for content_type in content_types:
        if re.match(content_type, ctype):
            return True
    return False


# Wrapper around requests.Session
# - log input and output data
# - use HTTP Basic auth if resource.basic_auth_username and resource.basic_auth_password exist
# - use client side certificate if resource.client_certificate (FileField) exists
# - verify server certificate CA if resource.trusted_certificate_authorities (FileField) exists
# - disable CA verification if resource.verify_cert (BooleanField) exists and is set
# - use a proxy for HTTP and HTTPS if resource.http_proxy exists

class Request(RequestSession):

    def __init__(self, *args, **kwargs):
        self.logger = kwargs.pop('logger')
        self.resource = kwargs.pop('resource', None)
        super(Request, self).__init__(*args, **kwargs)

    def request(self, method, url, **kwargs):
        cache_duration = kwargs.pop('cache_duration', None)
        invalidate_cache = kwargs.pop('invalidate_cache', False)
        params = kwargs.get('params', '')

        if self.resource:
            if 'auth' not in kwargs:
                username = getattr(self.resource, 'basic_auth_username', None)
                if username and hasattr(self.resource, 'basic_auth_password'):
                    kwargs['auth'] = (username, self.resource.basic_auth_password)
            if 'cert' not in kwargs:
                keystore = getattr(self.resource, 'client_certificate', None)
                if keystore:
                    kwargs['cert'] = keystore.path
            if 'verify' not in kwargs:
                trusted_certificate_authorities = getattr(self.resource,
                                                          'trusted_certificate_authorities',
                                                          None)
                if trusted_certificate_authorities:
                    kwargs['verify'] = trusted_certificate_authorities.path
                elif hasattr(self.resource, 'verify_cert'):
                    kwargs['verify'] = self.resource.verify_cert
            if 'proxies' not in kwargs:
                proxy = getattr(self.resource, 'http_proxy', None)
                if proxy:
                    kwargs['proxies'] = {'http': proxy, 'https': proxy}

        if method == 'GET' and cache_duration:
            cache_key = hashlib.md5('%r;%r' % (url, kwargs)).hexdigest()
            cache_content = cache.get(cache_key)
            if cache_content and not invalidate_cache:
                response = RequestResponse()
                response.raw = StringIO(cache_content.get('content'))
                response.headers = CaseInsensitiveDict(cache_content.get('headers', {}))
                response.status_code = cache_content.get('status_code')
                return response

        if settings.REQUESTS_PROXIES and 'proxies' not in kwargs:
            kwargs['proxies'] = settings.REQUESTS_PROXIES

        if 'timeout' not in kwargs:
            kwargs['timeout'] = settings.REQUESTS_TIMEOUT

        response = super(Request, self).request(method, url, **kwargs)

        if method == 'GET' and cache_duration and (response.status_code // 100 == 2):
            cache.set(cache_key, {
                'content': response.content,
                'headers': response.headers,
                'status_code': response.status_code,
            }, cache_duration)

        return response

    def send(self, request, **kwargs):
        try:
            response = super(Request, self).send(request, **kwargs)
        except Exception as exc:
            self.log_http_request(request, exception=exc)
            raise
        self.log_http_request(request, response)
        return response

    def log_http_request(self, request, response=None, exception=None):
        message = '%s %s' % (request.method, request.url)
        extra = {}
        extra['request_url'] = request.url
        if self.logger.level == 10:  # DEBUG
            extra['request_headers'] = dict(request.headers.items())
            if request.body:
                extra['request_payload'] = repr(request.body[:5000])
        log_function = self.logger.info
        if response is not None:
            message = message + ' (=> %s)' % response.status_code
            extra['response_status'] = response.status_code
            if self.logger.level == 10:  # DEBUG
                extra['response_headers'] = dict(response.headers.items())
                # log body only if content type is allowed
                if content_type_match(response.headers.get('Content-Type')):
                    content = response.content[:settings.LOGGED_RESPONSES_MAX_SIZE]
                    extra['response_content'] = repr(content)
            if response.status_code // 100 == 3:
                log_function = self.logger.warning
            elif response.status_code // 100 >= 4:
                log_function = self.logger.error
        elif exception:
            message = message + ' (=> %s)' % repr(exception)
            extra['response_exception'] = repr(exception)
            log_function = self.logger.error

        log_function(message, extra=extra)


class SOAPClient(Client):
    """Wrapper around zeep.Client

    resource muste have a wsdl_url and a requests attribute
    """
    def __init__(self, resource, **kwargs):
        transport = Transport(session=resource.requests, cache=InMemoryCache())
        wsdl_url = kwargs.pop('wsdl_url', None) or resource.wsdl_url
        super(SOAPClient, self).__init__(wsdl_url, transport=transport, **kwargs)


def export_site():
    '''Dump passerelle configuration (users, resources and ACLs) to JSON dumpable dictionnary'''
    d = {}
    d['apiusers'] = [apiuser.export_json() for apiuser in ApiUser.objects.all()]
    d['resources'] = resources = []
    for subclass in BaseResource.__subclasses__():
        if subclass._meta.abstract:
            continue
        for resource in subclass.objects.all():
            try:
                resources.append(resource.export_json())
            except NotImplementedError:
                break
    return d


def import_site(d, if_empty=False, clean=False, overwrite=False, import_users=False):
    '''Load passerelle configuration (users, resources and ACLs) from a dictionnary loaded from
       JSON
    '''
    d = d.copy()

    def is_empty():
        if import_users:
            if ApiUser.objects.count():
                return False

        for subclass in BaseResource.__subclasses__():
            if subclass._meta.abstract:
                continue
            if subclass.objects.count():
                return False
        return True

    if if_empty and not is_empty():
        return

    if clean:
        for subclass in BaseResource.__subclasses__():
            if subclass._meta.abstract:
                continue
            subclass.objects.all().delete()
        if import_users:
            ApiUser.objects.all().delete()

    with transaction.atomic():
        if import_users:
            for apiuser in d['apiusers']:
                ApiUser.import_json(apiuser, overwrite=overwrite)

        for resource in d['resources']:
            BaseResource.import_json(resource, overwrite=overwrite, import_users=import_users)


def batch(iterable, size):
    '''Batch an iterable as an iterable of iterables of at most size element
       long.
    '''
    sourceiter = iter(iterable)
    while True:
        batchiter = islice(sourceiter, size)
        # call next() at least one time to advance, if the caller does not
        # consume the returned iterators, sourceiter will never be exhausted.
        yield chain([batchiter.next()], batchiter)
