import pytest

from django.contrib.contenttypes.models import ContentType

from passerelle.apps.jsondatastore.models import JsonDataStore, JsonData
from passerelle.base.models import ApiUser, AccessRight


@pytest.fixture
def jsondatastore(db):
    datastore = JsonDataStore.objects.create(slug='foobar')
    api = ApiUser.objects.create(username='all', keytype='', key='')
    obj_type = ContentType.objects.get_for_model(datastore)
    AccessRight.objects.create(codename='can_access', apiuser=api,
                               resource_type=obj_type, resource_pk=datastore.pk)
    return datastore

@pytest.fixture
def jsondatastore2(db):
    datastore = JsonDataStore.objects.create(slug='foobar2')
    api = ApiUser.objects.create(username='all2', keytype='', key='')
    obj_type = ContentType.objects.get_for_model(datastore)
    AccessRight.objects.create(codename='can_access', apiuser=api,
                               resource_type=obj_type, resource_pk=datastore.pk)
    return datastore


def test_jsondatastore(app, jsondatastore, jsondatastore2):
    resp = app.get('/jsondatastore/foobar/data/')
    assert resp.json == {'data': [], 'err': 0}

    resp = app.post_json('/jsondatastore/foobar/data/create', params={'foo': 'bar'})
    uuid = resp.json['id']
    resp = app.get('/jsondatastore/foobar/data/')
    assert len(resp.json['data']) == 1
    assert resp.json['data'][0]['content'] == {'foo': 'bar'}

    resp = app.get('/jsondatastore/foobar/data/%s/' % uuid)
    assert resp.json['id'] == uuid
    assert resp.json['content'] == {'foo': 'bar'}

    resp = app.post_json('/jsondatastore/foobar/data/%s/' % uuid, params={'foo': 'bar2'})
    assert resp.json['id'] == uuid
    assert resp.json['content'] == {'foo': 'bar2'}

    resp = app.get('/jsondatastore/foobar/data/%s/' % uuid)
    assert resp.json['id'] == uuid
    assert resp.json['content'] == {'foo': 'bar2'}

    resp = app.patch_json('/jsondatastore/foobar/data/%s/' % uuid, params={'foo2': 'bar2'})
    assert resp.json['id'] == uuid
    assert resp.json['content'] == {'foo': 'bar2', 'foo2': 'bar2'}

    resp = app.post_json('/jsondatastore/foobar/data/%s/delete' % uuid)
    assert resp.json['err'] == 0

    resp = app.get('/jsondatastore/foobar/data/')
    assert len(resp.json['data']) == 0

    resp = app.post_json('/jsondatastore/foobar/data/create', params={'foo': 'bar'})
    uuid = resp.json['id']
    resp = app.get('/jsondatastore/foobar2/data/')
    assert len(resp.json['data']) == 0

    resp = app.post_json('/jsondatastore/foobar2/data/%s/delete' % uuid, status=404)
    assert resp.json['err'] == 1

def test_jsondatastore_name_id(app, jsondatastore):
    resp = app.get('/jsondatastore/foobar/data/')
    assert resp.json == {'data': [], 'err': 0}

    resp = app.post_json('/jsondatastore/foobar/data/create?name_id=xxx', params={'foo': 'bar'})
    uuid = resp.json['id']
    resp = app.get('/jsondatastore/foobar/data/')
    assert len(resp.json['data']) == 1
    assert resp.json['data'][0]['content'] == {'foo': 'bar'}

    resp = app.get('/jsondatastore/foobar/data/')
    assert len(resp.json['data']) == 1
    resp = app.get('/jsondatastore/foobar/data/?name_id=xxx')
    assert len(resp.json['data']) == 1
    resp = app.get('/jsondatastore/foobar/data/?name_id=yyy')
    assert len(resp.json['data']) == 0
    resp = app.get('/jsondatastore/foobar/data/?name_id=')
    assert len(resp.json['data']) == 0

    resp = app.post_json('/jsondatastore/foobar/data/create?name_id=yyy', params={'foo': 'bar'})
    uuid = resp.json['id']
    resp = app.get('/jsondatastore/foobar/data/')
    assert len(resp.json['data']) == 2
    resp = app.get('/jsondatastore/foobar/data/?name_id=yyy')
    assert len(resp.json['data']) == 1

    # try removing an existing entry with the wrong name id
    resp = app.post_json('/jsondatastore/foobar/data/%s/delete?name_id=zzz' % uuid, status=404)
    assert resp.json['err'] == 1

    # try removing an existing entry with a blank name id
    resp = app.post_json('/jsondatastore/foobar/data/%s/delete?name_id=' % uuid, status=404)
    assert resp.json['err'] == 1

    # and with the right one
    resp = app.post_json('/jsondatastore/foobar/data/%s/delete?name_id=yyy' % uuid)
    assert resp.json['err'] == 0

def test_jsondatastore_template(app, jsondatastore):
    jsondatastore.text_value_template = '{{foo}}'
    jsondatastore.save()
    resp = app.post_json('/jsondatastore/foobar/data/create', params={'foo': 'bar'})
    uuid = resp.json['id']
    resp = app.get('/jsondatastore/foobar/data/')
    assert len(resp.json['data']) == 1
    assert resp.json['data'][0]['text'] == 'bar'

def test_jsondatastore_get_by_attribute(app, jsondatastore):
    resp = app.post_json('/jsondatastore/foobar/data/create', params={'foo': 'bar'})
    uuid = resp.json['id']
    resp = app.post_json('/jsondatastore/foobar/data/create', params={'foo': 'bar2'})
    uuid2 = resp.json['id']

    resp = app.get('/jsondatastore/foobar/data/by/foo/', params={'value': 'bar'})
    assert resp.json['id'] == uuid

    resp = app.get('/jsondatastore/foobar/data/by/foo/', params={'value': 'bar2'})
    assert resp.json['id'] == uuid2

    resp = app.get('/jsondatastore/foobar/data/by/foo/', params={'value': 'bar3'})
    assert resp.json['err'] == 1

    resp = app.post_json('/jsondatastore/foobar/data/create?name_id=xxx', params={'foo': 'bar3'})
    uuid = resp.json['id']
    resp = app.post_json('/jsondatastore/foobar/data/create?name_id=yyy', params={'foo': 'bar3'})
    uuid2 = resp.json['id']

    resp = app.get('/jsondatastore/foobar/data/by/foo/', params={'value': 'bar3', 'name_id': 'xxx'})
    assert resp.json['id'] == uuid

    resp = app.get('/jsondatastore/foobar/data/by/foo/', params={'value': 'bar3', 'name_id': 'yyy'})
    assert resp.json['id'] == uuid2

    resp = app.get('/jsondatastore/foobar/data/by/foo/', params={'value': 'bar3', 'name_id': 'zzz'})
    assert resp.json['err'] == 1
